# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import copy
from unittest.mock import MagicMock
import pytest
from a2ui.schema.constants import (
    A2UI_OPEN_TAG,
    A2UI_CLOSE_TAG,
    VERSION_0_9,
    SURFACE_ID_KEY,
    CATALOG_COMPONENTS_KEY,
)
from a2ui.parser.constants import (
    MSG_TYPE_CREATE_SURFACE,
    MSG_TYPE_UPDATE_COMPONENTS,
    MSG_TYPE_DELETE_SURFACE,
    MSG_TYPE_DATA_MODEL_UPDATE,
)
from a2ui.schema.catalog import A2uiCatalog
from a2ui.parser.streaming import A2uiStreamParser
from a2ui.parser.response_part import ResponsePart


@pytest.fixture
def mock_catalog():
  s2c_schema = {
      "$schema": "https://json-schema.org/draft/2020-12/schema",
      "$id": "https://a2ui.org/specification/v0_9/server_to_client.json",
      "title": "A2UI Message Schema",
      "type": "object",
      "oneOf": [
          {"$ref": "#/$defs/CreateSurfaceMessage"},
          {"$ref": "#/$defs/UpdateComponentsMessage"},
          {"$ref": "#/$defs/UpdateDataModelMessage"},
          {"$ref": "#/$defs/DeleteSurfaceMessage"},
      ],
      "$defs": {
          "CreateSurfaceMessage": {
              "type": "object",
              "properties": {
                  "version": {"const": "v0.9"},
                  "createSurface": {
                      "type": "object",
                      "properties": {
                          "surfaceId": {
                              "type": "string",
                          },
                          "catalogId": {
                              "type": "string",
                          },
                          "theme": {
                              "type": "object",
                              "additionalProperties": True,
                          },
                      },
                      "required": ["surfaceId", "catalogId"],
                      "additionalProperties": False,
                  },
              },
              "required": ["version", "createSurface"],
              "additionalProperties": False,
          },
          "UpdateComponentsMessage": {
              "type": "object",
              "properties": {
                  "version": {"const": "v0.9"},
                  "updateComponents": {
                      "type": "object",
                      "properties": {
                          "surfaceId": {
                              "type": "string",
                          },
                          "root": {
                              "type": "string",
                          },
                          "components": {
                              "type": "array",
                              "minItems": 1,
                              "items": {"$ref": "catalog.json#/$defs/anyComponent"},
                          },
                      },
                      "required": ["surfaceId", "components"],
                      "additionalProperties": False,
                  },
              },
              "required": ["version", "updateComponents"],
              "additionalProperties": False,
          },
          "UpdateDataModelMessage": {
              "type": "object",
              "properties": {
                  "version": {"const": "v0.9"},
                  "updateDataModel": {
                      "type": "object",
                      "properties": {
                          "surfaceId": {
                              "type": "string",
                          },
                          "value": {"additionalProperties": True},
                      },
                      "required": ["surfaceId"],
                      "additionalProperties": False,
                  },
              },
              "required": ["version", "updateDataModel"],
              "additionalProperties": False,
          },
          "DeleteSurfaceMessage": {
              "type": "object",
              "properties": {
                  "version": {"const": "v0.9"},
                  "deleteSurface": {
                      "type": "object",
                      "properties": {"surfaceId": {"type": "string"}},
                      "required": ["surfaceId"],
                  },
              },
              "required": ["version", "deleteSurface"],
          },
      },
  }
  catalog_schema = {
      "catalogId": "test_catalog",
      "components": {
          "Container": {
              "type": "object",
              "allOf": [
                  {"$ref": "common_types.json#/$defs/ComponentCommon"},
                  {"$ref": "#/$defs/CatalogComponentCommon"},
              ],
              "properties": {
                  "component": {"const": "Container"},
                  "children": {
                      "type": "array",
                      "items": {"type": "string"},
                  },
              },
              "required": ["component", "children"],
          },
          "Card": {
              "type": "object",
              "allOf": [
                  {"$ref": "common_types.json#/$defs/ComponentCommon"},
                  {"$ref": "#/$defs/CatalogComponentCommon"},
              ],
              "properties": {
                  "component": {"const": "Card"},
                  "child": {"$ref": "common_types.json#/$defs/ComponentId"},
              },
              "required": ["component", "child"],
          },
          "Text": {
              "type": "object",
              "allOf": [
                  {"$ref": "common_types.json#/$defs/ComponentCommon"},
                  {"$ref": "#/$defs/CatalogComponentCommon"},
              ],
              "properties": {
                  "component": {"const": "Text"},
                  "text": {"$ref": "common_types.json#/$defs/DynamicString"},
              },
              "required": ["component", "text"],
          },
          "Column": {
              "type": "object",
              "allOf": [
                  {"$ref": "common_types.json#/$defs/ComponentCommon"},
                  {"$ref": "#/$defs/CatalogComponentCommon"},
              ],
              "properties": {
                  "component": {"const": "Column"},
                  "children": {"$ref": "common_types.json#/$defs/ChildList"},
              },
              "required": ["component", "children"],
          },
          "AudioPlayer": {
              "type": "object",
              "allOf": [
                  {"$ref": "common_types.json#/$defs/ComponentCommon"},
                  {"$ref": "#/$defs/CatalogComponentCommon"},
                  {
                      "type": "object",
                      "properties": {
                          "component": {"const": "AudioPlayer"},
                          "url": {"$ref": "common_types.json#/$defs/DynamicString"},
                          "description": {
                              "$ref": "common_types.json#/$defs/DynamicString"
                          },
                      },
                      "required": ["component", "url"],
                  },
              ],
          },
          "List": {
              "type": "object",
              "allOf": [
                  {"$ref": "common_types.json#/$defs/ComponentCommon"},
                  {"$ref": "#/$defs/CatalogComponentCommon"},
                  {
                      "type": "object",
                      "properties": {
                          "component": {"const": "List"},
                          "children": {"$ref": "common_types.json#/$defs/ChildList"},
                          "direction": {
                              "type": "string",
                              "enum": ["vertical", "horizontal"],
                          },
                      },
                      "required": ["component", "children"],
                  },
              ],
          },
          "Row": {
              "type": "object",
              "allOf": [
                  {"$ref": "common_types.json#/$defs/ComponentCommon"},
                  {"$ref": "#/$defs/CatalogComponentCommon"},
                  {
                      "type": "object",
                      "properties": {
                          "component": {"const": "Row"},
                          "children": {"$ref": "common_types.json#/$defs/ChildList"},
                      },
                      "required": ["component", "children"],
                  },
              ],
          },
      },
      "$defs": {
          "CatalogComponentCommon": {
              "type": "object",
              "properties": {"weight": {"type": "number"}},
          },
          "anyComponent": {
              "oneOf": [
                  {"$ref": "#/components/Container"},
                  {"$ref": "#/components/Card"},
                  {"$ref": "#/components/Text"},
                  {"$ref": "#/components/Column"},
                  {"$ref": "#/components/AudioPlayer"},
                  {"$ref": "#/components/List"},
                  {"$ref": "#/components/Row"},
              ],
              "discriminator": {"propertyName": "component"},
          },
      },
  }
  common_types_schema = {
      "$schema": "https://json-schema.org/draft/2020-12/schema",
      "$id": "https://a2ui.org/specification/v0_9/common_types.json",
      "title": "A2UI Common Types",
      "$defs": {
          "ComponentId": {
              "type": "string",
          },
          "AccessibilityAttributes": {
              "type": "object",
              "properties": {
                  "label": {
                      "$ref": "#/$defs/DynamicString",
                  }
              },
          },
          "Action": {"type": "object", "additionalProperties": True},
          "ComponentCommon": {
              "type": "object",
              "properties": {"id": {"$ref": "#/$defs/ComponentId"}},
              "required": ["id"],
          },
          "DataBinding": {"type": "object"},
          "DynamicString": {
              "anyOf": [{"type": "string"}, {"$ref": "#/$defs/DataBinding"}]
          },
          "DynamicValue": {
              "anyOf": [
                  {"type": "object"},
                  {"type": "array"},
                  {"$ref": "#/$defs/DataBinding"},
              ]
          },
          "DynamicNumber": {
              "anyOf": [{"type": "number"}, {"$ref": "#/$defs/DataBinding"}]
          },
          "ChildList": {
              "oneOf": [
                  {"type": "array", "items": {"$ref": "#/$defs/ComponentId"}},
                  {
                      "type": "object",
                      "properties": {
                          "componentId": {"$ref": "#/$defs/ComponentId"},
                          "path": {"type": "string"},
                      },
                      "required": ["componentId", "path"],
                      "additionalProperties": False,
                  },
              ]
          },
      },
  }
  return A2uiCatalog(
      version=VERSION_0_9,
      name="test_catalog",
      s2c_schema=s2c_schema,
      common_types_schema=common_types_schema,
      catalog_schema=catalog_schema,
  )


def _normalize_messages(messages):
  """Sorts components in messages for stable comparison."""
  # Support ResponsePart list by extracting a2ui_json
  res = []
  for m in messages:
    if isinstance(m, ResponsePart):
      if m.a2ui_json:
        if isinstance(m.a2ui_json, list):
          res.extend(copy.deepcopy(m.a2ui_json))
        else:
          res.append(copy.deepcopy(m.a2ui_json))
    else:
      res.append(copy.deepcopy(m))

  for msg in res:
    if MSG_TYPE_UPDATE_COMPONENTS in msg:
      payload = msg[MSG_TYPE_UPDATE_COMPONENTS]
      if CATALOG_COMPONENTS_KEY in payload:
        payload[CATALOG_COMPONENTS_KEY].sort(key=lambda x: x.get("id", ""))
  return res


def assertResponseContainsMessages(response, expected_messages):
  """Asserts that the response parts contain the expected messages."""
  assert _normalize_messages(response) == _normalize_messages(expected_messages)


def assertResponseContainsNoA2UI(response):
  assert len(response) == 0 or response[0].a2ui_json == None


def assertResponseContainsText(response, expected_text):
  """Asserts that the response parts contain the expected text."""
  assert any(
      (p.text if isinstance(p, ResponsePart) else p) == expected_text for p in response
  )


def test_add_msg_type_deduplication():
  parser = A2uiStreamParser()
  parser.add_msg_type(MSG_TYPE_UPDATE_COMPONENTS)
  parser.add_msg_type(MSG_TYPE_UPDATE_COMPONENTS)
  assert parser.msg_types == [MSG_TYPE_UPDATE_COMPONENTS]

  parser.add_msg_type(MSG_TYPE_CREATE_SURFACE)
  assert parser.msg_types == [MSG_TYPE_UPDATE_COMPONENTS, MSG_TYPE_CREATE_SURFACE]
  parser.add_msg_type(MSG_TYPE_UPDATE_COMPONENTS)
  assert parser.msg_types == [MSG_TYPE_UPDATE_COMPONENTS, MSG_TYPE_CREATE_SURFACE]


def test_streaming_msg_type_deduplication(mock_catalog):
  parser = A2uiStreamParser(catalog=mock_catalog)
  # 1. Send partial chunk that triggers sniffing
  chunk1 = (
      A2UI_OPEN_TAG
      + '[{"version": "v0.9", "updateComponents": {"surfaceId": "s1", "root": "root",'
      ' "components": [{"id": "root", "component": "Text", "text": "Hello"}'
  )
  parser.process_chunk(chunk1)

  assert MSG_TYPE_UPDATE_COMPONENTS in parser.msg_types
  assert parser.msg_types.count(MSG_TYPE_UPDATE_COMPONENTS) == 1

  # 2. Send the rest, which triggers handle_complete_object
  chunk2 = f', {{"id": "c1", "component": "Text", "text": "hi"}}]}}}} {A2UI_CLOSE_TAG}'
  parser.process_chunk(chunk2)

  # After completion, msg_types is reset
  assert not parser.msg_types


def test_v09_path_heuristic_relative_path(mock_catalog):
  """Tests that v0.9 allows relative paths (no leading slash)."""
  parser = A2uiStreamParser(catalog=mock_catalog)
  # Disable validation to avoid needing full catalog for this test
  parser._validator = None

  # 1. Create surface
  chunk_cs = (
      A2UI_OPEN_TAG
      + '[{"version": "v0.9", "createSurface": {"surfaceId": "s1", "catalogId": "c1"}}]'
      + A2UI_CLOSE_TAG
  )
  list(parser.process_chunk(chunk_cs))

  # 2. Update components with a relative path
  chunk_uc = (
      A2UI_OPEN_TAG
      + '[{"version": "v0.9", "updateComponents": {"surfaceId": "s1", "components":'
      ' [{"id": "root", "component": "Text", "text": {"path":'
      ' "some/relative/path"}}]}}]'
      + A2UI_CLOSE_TAG
  )

  messages = []
  for part in parser.process_chunk(chunk_uc):
    if part.a2ui_json:
      messages.extend(part.a2ui_json)

  assert len(messages) > 0
  comp = messages[0][MSG_TYPE_UPDATE_COMPONENTS]["components"][0]
  assert comp["text"]["path"] == "some/relative/path"


def test_v09_path_heuristic_absolute_path(mock_catalog):
  """Tests that v0.9 still supports absolute paths (leading slash)."""
  parser = A2uiStreamParser(catalog=mock_catalog)
  parser._validator = None

  # 1. Create surface
  chunk_cs = (
      A2UI_OPEN_TAG
      + '[{"version": "v0.9", "createSurface": {"surfaceId": "s1", "catalogId": "c1"}}]'
      + A2UI_CLOSE_TAG
  )
  list(parser.process_chunk(chunk_cs))

  # 2. Update components with an absolute path
  chunk_uc = (
      A2UI_OPEN_TAG
      + '[{"version": "v0.9", "updateComponents": {"surfaceId": "s1", "components":'
      ' [{"id": "root", "component": "Text", "text": {"path": "/absolute/path"}}]}}]'
      + A2UI_CLOSE_TAG
  )

  messages = []
  for part in parser.process_chunk(chunk_uc):
    if part.a2ui_json:
      messages.extend(part.a2ui_json)

  assert len(messages) > 0
  comp = messages[0][MSG_TYPE_UPDATE_COMPONENTS]["components"][0]
  assert comp["text"]["path"] == "/absolute/path"



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import yaml
import pytest


def _get_conformance_path(filename):
  return os.path.abspath(
      os.path.join(os.path.dirname(__file__), "../../../conformance", filename)
  )


def load_tests(filename):
  path = _get_conformance_path(os.path.join("suites", filename))
  with open(path, "r", encoding="utf-8") as f:
    return yaml.safe_load(f)


def get_conformance_cases(filename):
  cases = load_tests(filename)
  return [(case["name"], case) for case in cases]


# --- A2A Integration Conformance ---
cases_a2a_integration = get_conformance_cases("a2a_integration.yaml")


@pytest.mark.parametrize(
    "name, test_case",
    cases_a2a_integration,
    ids=[c[0] for c in cases_a2a_integration],
)
def test_a2a_integration_conformance(name, test_case):
  from a2ui.a2a.parts import create_a2ui_part, is_a2ui_part, get_a2ui_datapart
  from a2a.types import DataPart, Part
  from a2ui.a2a.extension import (
      get_a2ui_agent_extension,
      try_activate_a2ui_extension,
      _select_newest_a2ui_extension,
  )
  from unittest.mock import MagicMock
  from a2a.server.agent_execution import RequestContext

  action = test_case["action"]
  args = test_case.get("args", {})

  if action == "create_a2ui_part":
    data = args["data"]
    part = create_a2ui_part(data)
    assert is_a2ui_part(part)
    expect = test_case["expect"]
    data_part = get_a2ui_datapart(part)
    assert data_part.metadata.get("mimeType") == expect["mime_type"]

  elif action == "is_a2ui_part":
    mime_type = args["mime_type"]
    part = Part(root=DataPart(data={}, metadata={"mimeType": mime_type}))
    result = is_a2ui_part(part)
    assert result == test_case["expect"]

  elif action == "get_extension":
    version = args["version"]
    accepts_inline_catalogs = args.get("accepts_inline_catalogs")
    supported_catalog_ids = args.get("supported_catalog_ids")

    kwargs = {}
    if accepts_inline_catalogs is not None:
      kwargs["accepts_inline_catalogs"] = accepts_inline_catalogs
    if supported_catalog_ids is not None:
      kwargs["supported_catalog_ids"] = supported_catalog_ids

    ext = get_a2ui_agent_extension(version, **kwargs)
    expect = test_case["expect"]
    assert ext.uri == expect["uri"]
    if expect["params"] is None:
      assert ext.params is None
    else:
      assert ext.params == expect["params"]

  elif action == "try_activate":
    requested = args["requested"]
    advertised = args["advertised"]

    context = MagicMock(spec=RequestContext)
    context.requested_extensions = requested
    context.add_activated_extension = MagicMock()

    card = MagicMock()
    extensions = []
    for uri in advertised:
      ext = MagicMock()
      ext.uri = uri
      extensions.append(ext)
    card.capabilities.extensions = extensions

    result_version = try_activate_a2ui_extension(context, card)
    expect = test_case["expect"]

    if expect["activated"] is None:
      assert result_version is None
      context.add_activated_extension.assert_not_called()
    else:
      assert result_version == expect["version"]
      context.add_activated_extension.assert_called_once_with(expect["activated"])

  elif action == "select_newest":
    requested = args["requested"]
    advertised = args["advertised"]
    result = _select_newest_a2ui_extension(requested, advertised)
    expect = test_case["expect"]
    assert result == expect["newest"]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Optional, List

from a2a.server.agent_execution import RequestContext
from a2a.types import AgentCard, AgentExtension

logger = logging.getLogger(__name__)

A2UI_EXTENSION_BASE_URI = "https://a2ui.org/a2a-extension/a2ui"
AGENT_EXTENSION_SUPPORTED_CATALOG_IDS_KEY = "supportedCatalogIds"
AGENT_EXTENSION_ACCEPTS_INLINE_CATALOGS_KEY = "acceptsInlineCatalogs"


def get_a2ui_agent_extension(
    version: str,
    accepts_inline_catalogs: bool = False,
    supported_catalog_ids: List[str] = [],
) -> AgentExtension:
  """Creates the A2UI AgentExtension configuration.

  Args:
      version: The version of the A2UI extension to use.
      accepts_inline_catalogs: Whether the agent accepts inline catalogs.
      supported_catalog_ids: All pre-defined catalogs the agent is known to support.

  Returns:
      The configured A2UI AgentExtension.
  """
  params = {}
  if accepts_inline_catalogs:
    params[AGENT_EXTENSION_ACCEPTS_INLINE_CATALOGS_KEY] = (
        True  # Only set if not default of False
    )

  if supported_catalog_ids:
    params[AGENT_EXTENSION_SUPPORTED_CATALOG_IDS_KEY] = supported_catalog_ids

  return AgentExtension(
      uri=f"{A2UI_EXTENSION_BASE_URI}/v{version}",
      description="Provides agent driven UI using the A2UI JSON format.",
      params=params if params else None,
  )


def _agent_extensions(agent_card: AgentCard) -> List[str]:
  """Returns the A2UI extension URIs supported by the agent."""
  extensions = []
  if (
      agent_card
      and hasattr(agent_card, "capabilities")
      and agent_card.capabilities
      and hasattr(agent_card.capabilities, "extensions")
      and agent_card.capabilities.extensions
  ):
    for ext in agent_card.capabilities.extensions:
      if ext.uri and ext.uri.startswith(A2UI_EXTENSION_BASE_URI):
        extensions.append(ext.uri)
  return extensions


def _requested_a2ui_extensions(context: RequestContext) -> List[str]:
  """Returns the A2UI extension URIs requested by the client."""
  requested_extensions = []
  if hasattr(context, "requested_extensions") and context.requested_extensions:
    requested_extensions.extend([
        ext
        for ext in context.requested_extensions
        if isinstance(ext, str) and ext.startswith(A2UI_EXTENSION_BASE_URI)
    ])

  if (
      hasattr(context, "message")
      and context.message
      and hasattr(context.message, "extensions")
      and context.message.extensions
  ):
    requested_extensions.extend([
        ext
        for ext in context.message.extensions
        if isinstance(ext, str) and ext.startswith(A2UI_EXTENSION_BASE_URI)
    ])

  return requested_extensions


def _select_newest_a2ui_extension(
    requested_extensions: List[str], agent_advertised_extensions: List[str]
) -> Optional[str]:
  """Selects the newest A2UI extension URI from the matched extensions."""
  matched_extensions = [
      uri for uri in requested_extensions if uri in agent_advertised_extensions
  ]
  if not matched_extensions:
    return None

  def _version_key(uri: str) -> tuple:
    version_str = uri.replace(f"{A2UI_EXTENSION_BASE_URI}/v", "")
    from packaging.version import parse as parse_version

    return parse_version(version_str)

  return max(matched_extensions, key=_version_key)


def try_activate_a2ui_extension(
    context: RequestContext, agent_card: AgentCard
) -> Optional[str]:
  """Activates the A2UI extension if requested.

  Args:
      context: The request context to check.
      agent_card: The agent card to check supported extensions.

  Returns:
      The version string of the activated A2UI extension, or None if not activated.
  """
  requested_extensions = _requested_a2ui_extensions(context)
  if not requested_extensions:
    return None

  agent_advertised_extensions = _agent_extensions(agent_card)
  if not agent_advertised_extensions:
    return None

  selected_uri = _select_newest_a2ui_extension(
      requested_extensions, agent_advertised_extensions
  )
  if selected_uri:
    context.add_activated_extension(selected_uri)
    return selected_uri.replace(f"{A2UI_EXTENSION_BASE_URI}/v", "")

  return None
